#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 10 19:00:41 2020

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import pandas as pd
from scipy.stats import anderson
from statsmodels.tsa.stattools import adfuller
from statsmodels.stats.stattools import durbin_watson 
from statsmodels.stats.diagnostic import acorr_ljungbox

#::: allesfitter modules
from . import config
from .general_output import logprint

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




def alles_anderson(residuals):
    '''
    Parameters
    ----------
    residuals : array of float
        The residuals after the fit of model+baseline+stellar_var.

    Returns
    -------
    isNormal : bool
        True if the residuals are normally distributed, False otherwise.
    
    Outputs
    -------
    It also prints the statstics and conclusions.
    
    Sauces
    ------
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.anderson.html
    https://www.statology.org/anderson-darling-test-python/
    '''
    logprint('Anderson-Darling Test')
    logprint('---------------------')
    logprint('This tests the null hypothesis that the residuals follows a normal distribution.') 
    statistic, critical_values, significance_levels = anderson(residuals)
    logprint('Test statistic\t\t', statistic)
    logprint('Critical values\t\t', critical_values)
    logprint('Significance levels\t', significance_levels/100.)
    logprint('Does the null hypotheses hold at a significance level of...')
    for cv, sl in zip(critical_values, significance_levels):
        logprint('...', sl/100., '\t\t', statistic < cv)
    isNormal = all(statistic < critical_values)
    if isNormal:
        logprint('The null hypothesis cannot be rejected.')
        logprint('In simple words: your residuals look good.')
    else:
        logprint('The null hypothesis is rejected at some significance levels.')
        logprint('In simple words: there might still be some structure in your residuals.')
    logprint('\n')
    
    return isNormal
        


def alles_adfuller(residuals):
    '''
    Parameters
    ----------
    residuals : array of float
        The residuals after the fit of model+baseline+stellar_var.

    Returns
    -------
    isStationary : bool
        True if the residuals are stationary, False otherwise.
    
    Outputs
    -------
    It also prints the statstics and conclusions.
    
    Sauces
    ------
    https://www.hackdeploy.com/augmented-dickey-fuller-test-in-python/
    '''
    class StationarityTests:
        def __init__(self, significance=.05):
            self.SignificanceLevel = significance
            self.pValue = None
            self.isStationary = None
            
        def ADF_Stationarity_Test(self, timeseries, printResults = True):
            logprint('Augmented Dickey-Fuller Test')
            logprint('----------------------------')
            logprint('This tests the null hypothesis that the residuals show non-stationarity (trends).') 
            adfTest = adfuller(timeseries, autolag='AIC')
            self.pValue = adfTest[1]
            if (self.pValue<self.SignificanceLevel):
                self.isStationary = True
            else:
                self.isStationary = False
            if printResults:
                dfResults = pd.Series(adfTest[0:4], index=['Test Statistic','P-Value','# Lags Used','# Observations Used'])
                for key,value in adfTest[4].items():
                    dfResults['Critical Value (%s)'%key] = value
                logprint(dfResults.to_string())
    
    sTest = StationarityTests()
    sTest.ADF_Stationarity_Test(residuals, printResults = True)
    logprint("Is the time series stationary? {0}".format(sTest.isStationary))
    if not sTest.isStationary:
        logprint('In simple words: there might still be some structure in your residuals.')
    else:
        logprint('In simple words: your residuals look good.')
    logprint('\n')
    
    return sTest.isStationary
        
    

def alles_durbin(residuals):
    '''
    Parameters
    ----------
    residuals : array of float
        The residuals after the fit of model+baseline+stellar_var.

    Returns
    -------
    isUncorrelated : bool
        True if the residuals are not correlated, False otherwise.
    
    Outputs
    -------
    It also prints the statstics and conclusions.
    
    Sauces
    ------
    https://www.statology.org/durbin-watson-test-python/
    '''
    statistic = durbin_watson(residuals) 
    isUncorrelated = (statistic>=1.5) & (statistic<=2.5)
    logprint('Durbin-Watson Test')
    logprint('------------------')
    logprint('This tests the null hypothesis that there is no correlation among the residuals.')
    logprint('Test statistic:', statistic)
    logprint('The test statistic is approximately equal to 2*(1-r) where r is the sample autocorrelation of the residuals. Interpretation:')
    logprint('\t< 1.5 suggests a positive correlation')
    logprint('\t1.5 to 2.5 suggests no correlation')
    logprint('\t> 2.5 suggests a negative correlation')
    if isUncorrelated:
        logprint('No signs of a correlation.')
        logprint('In simple words: your residuals look good.')
    elif statistic<1.5:
        logprint('Signs of a positive correlation.')
        logprint('In simple words: there might still be some structure in your residuals.')
    elif statistic>2.5:
        logprint('Signs of a negative correlation.')
        logprint('In simple words: there might still be some structure in your residuals.')
    logprint('\n')

    return isUncorrelated 

    

def alles_ljung(residuals):
    '''
    Parameters
    ----------
    residuals : array of float
        The residuals after the fit of model+baseline+stellar_var.

    Returns
    -------
    isUncorrelated : bool
        True if the residuals are not correlated, False otherwise.
    
    Outputs
    -------
    It also prints the statstics and conclusions.
    
    Sauces
    ------
    https://www.statology.org/ljung-box-test-python/
    https://www.statsmodels.org/dev/generated/statsmodels.stats.diagnostic.acorr_ljungbox.html?highlight=ljung
    '''
    
    logprint('Ljung-Box Test')
    logprint('--------------')
    logprint('This tests the null hypothesis that there is no correlation among the residuals.')
    
    df = acorr_ljungbox(residuals, lags=[1,5,10,15,20], return_df=True)
    df.reset_index(inplace=True)
    df = df.rename(columns = {'index':'lag'})
    logprint('Does the null hypotheses hold at a significance level of...')
    df['0.15'] = df['lb_pvalue'] > 0.15 #if Ture, the null hypothesis cannot be rejected at this significance level
    df['0.1'] = df['lb_pvalue'] > 0.10 #if Ture, the null hypothesis cannot be rejected at this significance level
    df['0.05'] = df['lb_pvalue'] > 0.05 #if Ture, the null hypothesis cannot be rejected at this significance level
    df['0.025'] = df['lb_pvalue'] > 0.025 #if Ture, the null hypothesis cannot be rejected at this significance level
    df['0.01'] = df['lb_pvalue'] > 0.01 #if Ture, the null hypothesis cannot be rejected at this significance level
    isUncorrelated = all(df['0.15']==True) & all(df['0.1']==True) & all(df['0.05']==True) & all(df['0.025']==True) & all(df['0.01']==True)
    logprint(df.to_string(index=False))
    if isUncorrelated:
        logprint('The null hypothesis cannot be rejected.')
        logprint('In simple words: your residuals look good.')
    else:
        logprint('The null hypothesis is rejected at some significance levels.')
        logprint('In simple words: there might still be some structure in your residuals.')
    logprint('\n')

    return isUncorrelated 
    
    
    
def residual_stats(residuals):
    '''
    Parameters
    ----------
    residuals : array of float
        The residuals after the fit of model+baseline+stellar_var.
    typ : str
        'mcmc' or 'ns', just givs the name to the output file.

    Returns
    -------
    None.
    
    Outputs
    -------
    Prints the statstics and conclusions.
    '''
    residuals = residuals[ ~np.isnan(residuals) & np.isfinite(residuals) ] #remove all evil
    logprint("\nPerforming diagnostic tests on the fit's residuals...\n")
    passed_anderson = alles_anderson(residuals)
    passed_adfuller = alles_adfuller(residuals)
    passed_durbin = alles_durbin(residuals)
    try:
        passed_ljung = alles_ljung(residuals)
    except:
        logprint('Ljung-Box Test crashed.')
        passed_ljung = '(crashed)'
    logprint('Summary')
    logprint('-------')
    logprint('Test                    Passed?')
    logprint('Anderson-Darling       ',passed_anderson)
    logprint('Augmented Dickey-Fuller',passed_adfuller)
    logprint('Durbin-Watson          ',passed_durbin)
    logprint('Ljung-Box              ',passed_ljung)
    
    
    
# if __name__ == "__main__":
#     residuals = np.random.normal(size=1000)
#     alles_residual_stats(residuals)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May  9 10:01:30 2019

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os
import numpy as np
import matplotlib.pyplot as plt

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

    
    
    
###############################################################################
#::: fct to check if the right signal was found
###############################################################################
def is_multiple_of(a, b, tolerance=0.05):
    a = np.float(a)
    b = np.float(b) 
    result = a % b
    return (abs(result/b) <= tolerance) or (abs((b-result)/b) <= tolerance)


def is_detected(inj_period, inj_epoch, tls_period, tls_epoch, tls_SNR, SNR_threshold):
    a = is_multiple_of(tls_period, inj_period/2.) #check if it is a multiple of half the period to within 5%
    b = ((abs(tls_epoch-inj_epoch)%inj_period) < 0.05) or ((abs(abs(tls_epoch-inj_epoch)%inj_period)-0.5) < 0.05) #check if the epoch's match to within 5% of the period (also allow half-period offsets)
    c = tls_SNR>=SNR_threshold
#    right_epoch = False
#    for tt in results.transit_times:
#        for i in range(-5,5):
#            right_epoch = right_epoch or (np.abs(tt-epoch+i*period) < (1./24.)) #check if any epochs matches to within 1 hour
             
#    right_depth   = (np.abs(np.sqrt(1.-results.depth)*rstar - rplanet)/rplanet < 0.05) #check if the depth matches
                
    if a*b*c:
        return True
    else:
        return False
                    
    
    
def is_detected_list(results, SNR_threshold):
    detected = [ is_detected(results['inj_period'][i], results['inj_epoch'][i], results['tls_period'][i], results['tls_epoch'][i], results['tls_SNR'][i], SNR_threshold) for i in range(len(results))]
    return np.array(detected)
    
    

###############################################################################
#::: plot
###############################################################################
def irplot(fname, SNR_threshold, period_bins=None, rplanet_bins=None, options=None):
    
    #::: handle inputs
    if options is None: options = {}
    if 'cmap' not in options: options['cmap'] = 'Blues_r'
    if 'logfname' not in options: options['logfname'] = 'injection_recovery_test.csv'
    if 'outdir' not in options: options['outdir'] = ''
    if len(options['outdir'])>0 and not os.path.exists(options['outdir']): os.makedirs(options['outdir'])
    
    #::: load the files and check which TLS detection matches an injection; 
    #::: note that one injection will have multiple TLS detections (due not false positives)
    results = np.genfromtxt(fname, delimiter=',', dtype=None, names=True)
    inj_periods = np.atleast_1d(results['inj_period'])
    inj_rplanets = np.atleast_1d(results['inj_rplanet'])
    try: 
        inj_depths = np.atleast_1d(results['inj_depth'])
    except:
        inj_depths = np.nan*inj_rplanets
    # tls_periods = np.atleast_1d(results['tls_period'])
    # tls_SNRs = np.atleast_1d(results['tls_SNR'])
    detected = is_detected_list(results, SNR_threshold)
    # print(detected)
    
    #::: now boil it down to unique injections and see whether any TLS detection matched it
    # period = np.unique(inj_periods)
    # rplanet = np.unique(inj_rplanets)
    period = []
    rplanet = []
    depth = []
    found = []
    for p in np.unique(inj_periods):
        for r in np.unique(inj_rplanets):
            period.append(p)
            rplanet.append(r)
            ind = np.where( (inj_periods==p) & (inj_rplanets==r) )[0]
            f = any( detected[ind] )
            found.append(f)
            depth.append(inj_depths[ind][0])
            # print(p,r,ind,f)
    period = np.array(period)
    rplanet = np.array(rplanet)
    depth = np.array(depth)
    found = np.array(found)
    
    
    
    ###############################################################################
    #::: match
    ###############################################################################
    def get_inj_depth_from_inj_rplanet(r):
        plt.figure()
        plt.plot(np.sort(rplanet), np.sort(depth))
        return np.interp( r, np.sort(rplanet), np.sort(depth) )
    
    
    def get_inj_rplanet_from_inj_depth(d):
        d = np.array(d)
        d[ d>np.nanmax(depth) ] = np.nan
        d[ d<np.nanmin(depth) ] = np.nan
        plt.figure()
        plt.plot(np.sort(rplanet), np.sort(depth))
        return np.interp( d, np.sort(depth), np.sort(rplanet) )
        
    
    
    ###############################################################################
    #::: scatter plot
    ###############################################################################
    if ( len(np.unique(inj_periods)) * len(np.unique(inj_periods)) ) < 100:
        print('\nCreating scatter plot (too few sampled for 2D histogram plot).\n')
        fig, ax = plt.subplots(figsize=(5,5))
        ax.scatter(period, rplanet, c=found, s=100/np.log10(len(period)), cmap=options['cmap'], edgecolors='b')
        ax.set(xlabel='Period (days)', ylabel='Radius '+r'$(R_\oplus)$')
        ax.text(0.5,1.05,'SNR>'+str(SNR_threshold)+' | blue: not recovered | white: recovered',ha='center',va='center',transform=ax.transAxes)
        try: fig.savefig( os.path.join(options['outdir'],options['logfname'].split('.')[0]+'_snr'+str(SNR_threshold)+'_scatter.pdf'), bbox_inches='tight') #some matplotlib versions crash when saving pdf...
        except: fig.savefig( os.path.join(options['outdir'],options['logfname'].split('.')[0]+'_snr'+str(SNR_threshold)+'_scatter.jpg'), bbox_inches='tight') #some matplotlib versions need pillow for jpg (conda install pillow)...
        plt.close(fig)
        # err
    
    
    
    ###############################################################################
    #::: histogram (normed)
    ###############################################################################
    if ( len(np.unique(inj_periods)) * len(np.unique(inj_periods)) ) >= 100:
        print('\nCreating 2D histogram plot (too many samples for scatter plot).\n')
        if (period_bins is not None) and (rplanet_bins is not None):
            bins = [period_bins, rplanet_bins]
        else:
            bins = [np.histogram_bin_edges(period, bins='auto'), np.histogram_bin_edges(rplanet, bins='auto')]
        h1,x,y = np.histogram2d(period[found==1], rplanet[found==1], bins=bins)
        h2,x,y = np.histogram2d(period[found==0], rplanet[found==0], bins=bins)
        normed_hist = (100.*h1/(h1+h2))
        
        fig, ax = plt.subplots(figsize=(6.8,5))
        X, Y = np.meshgrid(x, y)
        im = ax.pcolormesh(X, Y, normed_hist.T, cmap=options['cmap'], vmin=0, vmax=100);
        # im = plt.imshow(normed_hist.T, origin='lower', extent=(x[0], x[-1], y[0], y[-1]), interpolation='none', aspect='auto', cmap='jet', vmin=0, vmax=100, rasterized=True)
        plt.colorbar(im, label='Recovery rate (%)', pad=0.12)
        ax.set(xlabel='Injected period (days)', ylabel=r'Injected radius (R$_\oplus$)', title='SNR>'+str(SNR_threshold))
        # change_font(ax)
        
        ax1 = ax.twinx()
        y0, y1 = ax.get_ylim()
        ax1.set(ylabel = 'Observed depth (ppt)',
                # ylim = get_inj_depth_from_inj_rplanet(ax.get_ylim()))
                ylim = ax.get_ylim(), #in Rearth
                yticks = get_inj_rplanet_from_inj_depth([0.1,0.5,1,2,5,10,15,20,50,100]), #tick positions in rearth
                yticklabels = [0.1,0.5,1,2,5,10,15,20,50,100]) #labels in ppt
                # yticks = [0,0.25,0.5,0.75,1],
                # yticklabels = [ np.format_float_positional( get_inj_depth_from_inj_rplanet(y0+(y1-y0)*item), 1 ) for item in [0,0.25,0.5,0.75,1]])
        
        try: fig.savefig( os.path.join(options['outdir'],options['logfname'].split('.')[0]+'_snr'+str(SNR_threshold)+'_hist.pdf'), bbox_inches='tight') #some matplotlib versions crash when saving pdf...
        except: fig.savefig( os.path.join(options['outdir'],options['logfname'].split('.')[0]+'_snr'+str(SNR_threshold)+'_hist.jpg'), bbox_inches='tight') #some matplotlib versions need pillow for jpg (conda install pillow)...   
        plt.close(fig)
    
    
    
    ###############################################################################
    #::: pyplot histograms (total counts)
    ###############################################################################
    #fig, ax = plt.subplots(figsize=(6.5,5))
    #h1,x,y,im = plt.hist2d(period[found==1], rplanet[found==1], bins=bins, cmap='Blues_r')
    #plt.colorbar(im, label='Recovery rate (%)')
    #plt.xlabel('Injected period (days)')
    #plt.ylabel(r'Injected radius (R$_\oplus$)')
    #change_font(ax)
    
    
    #fig, ax = plt.subplots(figsize=(6.5,5))
    #h2,x,y,im = plt.hist2d(period[found==0], rplanet[found==0], bins=bins, cmap='Blues_r')
    #plt.colorbar(im, label='Recovery rate (%)')
    #plt.xlabel('Injected period (days)')
    #plt.ylabel(r'Injected radius (R$_\oplus$)')
    #change_font(ax)
    
    
    
    
    ###############################################################################
    #::: kdeplots
    ###############################################################################
    # fig, ax = plt.subplots(figsize=(6.5,5))
    # ax = sns.kdeplot(period[found==0], rplanet[found==0], shade=True, cmap='Blues', cbar=True, alpha=0.5)
    # ax = sns.kdeplot(period[found==1], rplanet[found==1], shade=True, cmap='Blues_r', cbar=True)
    # ax.set(xlim=[0,120], ylim=[0.8,4])
    # plt.xlabel('Injected period (days)')
    # plt.ylabel(r'Injected radius (R$_\oplus$)')
    # change_font(ax)
    
    
    #fig, ax = plt.subplots(figsize=(6.5,5))
    #ax = sns.kdeplot(period[found==0], rplanet[found==0], shade=True, cmap='Blues', cbar=True)
    #ax.set(xlim=[15,85], ylim=[0.8,2.0])
    #plt.xlabel('Injected period (days)')
    #plt.ylabel(r'Injected radius (R$_\oplus$)')
    #change_font(ax)
    
    
    
    
    ###############################################################################
    #::: others
    ###############################################################################
    # plt.figure(figsize=(5,5))
    # z = found.reshape(len(np.unique(period)), len(np.unique(rplanet)))
    # plt.imshow(z.T, origin='lower', extent=(np.amin(period), np.amax(period), np.amin(rplanet), np.amax(rplanet)), aspect='auto', interpolation='gaussian', filterrad=5, cmap='Blues_r')
    # plt.xlabel('Period (days)')
    # plt.ylabel(r'Radius (R$_\oplus$)')
    
    #fig, ax = plt.subplots(figsize=(6.5,5))
    #plt.tricontourf(period, rplanet, found, cmap='Blues_r')
    #plt.xlabel('Injected period (days)')
    #plt.ylabel(r'Injected radius (R$_\oplus$)')
    
    
    # grid_x, grid_y = np.mgrid[np.amin(period):np.amax(period):100j, np.amin(rplanet):np.amax(rplanet):100j]
    # grid_z = griddata((period, rplanet), found*100, (grid_x, grid_y), method='linear')
    # fig, ax = plt.subplots(figsize=(6.5,5))
    # im = plt.imshow(grid_z.T, origin='lower', extent=(np.amin(period), np.amax(period), np.amin(rplanet), np.amax(rplanet)), interpolation='none', aspect='auto', cmap='Blues_r', rasterized=True, vmin=0, vmax=100)
    # plt.colorbar(im, label='Recovery rate (%)')
    # plt.xlabel('Injected period (days)')
    # plt.ylabel(r'Injected radius (R$_\oplus$)')
    #change_font(ax)
    #    
    #plt.savefig('injected_transit_search.pdf', bbox_inches='tight')       




###############################################################################
#::: run
###############################################################################
# plot('TIC_269701147.csv')







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan 31 15:50:31 2020

@author: 
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import os
import warnings

#::: specific modules
try:
    from wotan import flatten
except ImportError:
    pass

#::: my modules
import allesfitter
from allesfitter.lightcurves import eclipse_width_smart
from allesfitter.exoworlds_rdx.lightcurves.index_transits import index_transits, get_tmid_observed_transits
from allesfitter.plotting import fullplot, brokenplot, chunkplot, monthplot, tessplot


    

###############################################################################
#::: prepare TTV fit (if chosen)
###############################################################################
def prepare_ttv_fit(datadir, style='fullplot'):
    '''
    Inputs:
    -------
    datadir : str
        the working directory for allesfitter; must contain all the data files; output directories and files will also be created inside datadir
    style : str
        chose between 'fullplot', 'monthplot', and 'tessplot'; this defines how the plot looks
        default is 'fullplot'
        
    Outputs:
    --------
    None
    
    Notes:
    ------
    This function must be run *after* reduce_phot_data()
    Throughout, we use fast_fit_width as the approximation for the transit window
    '''
    
    #----------------------------------------------------------------------
    #::: setup
    #----------------------------------------------------------------------
    alles = allesfitter.allesclass(datadir)
    window = alles.settings['fast_fit_width']
    # colors = plt.rcParams["axes.prop_cycle"].by_key()["color"] #the default sequence of plot colors (for the companions)
    if not os.path.exists( os.path.join(datadir,'ttv_preparation') ): os.makedirs(os.path.join(datadir,'ttv_preparation'))
    with open(os.path.join(datadir,'ttv_preparation','ttv_initial_guess_params.csv'),'w') as f:
        f.write('')
                
        
    #----------------------------------------------------------------------
    #::: get the combined, full data from all instruments
    #----------------------------------------------------------------------
    time_combined = []
    flux_combined = []
    for inst in alles.settings['inst_phot']:
        time_combined += list(alles.fulldata[inst]['time'])
        flux_combined += list(alles.fulldata[inst]['flux'])
    ind_sort = np.argsort(time_combined)
    time_combined = np.array(time_combined)[ind_sort]
    flux_combined = np.array(flux_combined)[ind_sort]
    
    
    #----------------------------------------------------------------------
    #::: get eclipse widths per companion
    #----------------------------------------------------------------------
    eclipse_width = {}
    for companion in alles.settings['companions_phot']:
        alles.initial_guess_params_median[companion+'_epoch']
        eclipse_width[companion] = eclipse_width_smart(alles.initial_guess_params_median[companion+'_period'], 
                                                        alles.initial_guess_params_median[companion+'_rr'], 
                                                        alles.initial_guess_params_median[companion+'_rsuma'], 
                                                        alles.initial_guess_params_median[companion+'_cosi'], 
                                                        alles.initial_guess_params_median[companion+'_f_s'], 
                                                        alles.initial_guess_params_median[companion+'_f_c'])[0]
    
    
    #----------------------------------------------------------------------
    #::: loop over all photometric companions
    #----------------------------------------------------------------------
    for companion in alles.settings['companions_phot']:
        with open(os.path.join(datadir,'ttv_preparation','ttv_initial_guess_params.csv'),'a') as f:
            f.write('#TTV companion '+companion+',,,,,\n')
        
        
        #----------------------------------------------------------------------
        #::: compute tmid and a ttv_guess, and make per-transit-plots per companion
        #----------------------------------------------------------------------
        tmid_estimates = []
        tmid_linear_predictions = get_tmid_observed_transits(time_combined, 
                                                             alles.initial_guess_params_median[companion+'_epoch'], 
                                                             alles.initial_guess_params_median[companion+'_period'],
                                                             window)
        N = len(tmid_linear_predictions)
        fig, axes = plt.subplots(N, 1, figsize=(6,4*N), sharey=True, tight_layout=True)
        
        for i, tmid1 in enumerate(tmid_linear_predictions):
            
            #::: estimate the observed tranist midtime by computing the minimum of the data around the linearly predicted transit midtime
            ind_tr1 = np.where((time_combined >= (tmid1 - window/2.)) & (time_combined <= (tmid1 + window/2.)))[0]
            tr_times = time_combined[ind_tr1]
            tr_flux = flux_combined[ind_tr1]
            t_exp = np.median(np.diff(tr_times))
            N_points_in_eclipse = int(eclipse_width[companion]/t_exp)
            try:
                trend = flatten(tr_times, tr_flux, window_length=eclipse_width[companion]/2., method='biweight', return_trend=True)[1]
                tmid2 = np.median( tr_times[ np.argsort(trend)[0:int(N_points_in_eclipse/2.)] ] ) # the tmid estimated from the observations as the minimum of the data
            except:
                warnings.warn('Install wotan for improved performance of prepare_ttv_fit().')
                trend = None
                tmid2 = np.median( tr_times[ np.argsort(tr_times)[0:int(N_points_in_eclipse/2.)] ] ) # the tmid estimated from the observations as the minimum of the data
            ttv_guess = tmid2 - tmid1 # a guess for the TTV
            tmid_estimates.append(tmid2)
            
            #::: plot this per transit
            ax = axes[i]
            ax.plot(tr_times, tr_flux, marker='.', ls='none', color=alles.settings[companion+'_color'], rasterized=True)
            if trend is not None: ax.plot(tr_times, trend, 'r-')
            ax.axvline(tmid1,c='grey',ls='--',label='linear prediction')
            ax.axvline(tmid2,c='r',ls='--',label='flux minimum')
            ax.set(xlabel='Time', ylabel='Flux', xlim=[tmid1-window/2., tmid1+window/2.])
            ax.text(0.95,0.95,'Transit '+str(i+1), va='top', ha='right', transform=ax.transAxes)
            with open(os.path.join(datadir,'ttv_preparation','ttv_initial_guess_params.csv'),'a') as f:
               f.write(companion+'_ttv_transit_'+str(i+1)+','+np.format_float_positional(ttv_guess,4)+',1,uniform '+np.format_float_positional(ttv_guess-0.01,4)+' '+np.format_float_positional(ttv_guess+0.01,4)+',TTV$_\mathrm{'+companion+';'+str(i+1)+'}$,d\n')
        
        axes[0].legend()
        fig.savefig(os.path.join(datadir,'ttv_preparation','ttv_preparation_'+companion+'_per_transit.pdf'), bbox_inches='tight')
        plt.close(fig)
         
        tmid_estimates = np.array(tmid_estimates)
        
        
        #----------------------------------------------------------------------
        #::: ttv guess 0-C plot
        #----------------------------------------------------------------------
        nr = np.array([ int(np.round( (t-tmid_estimates[0]) / alles.initial_guess_params_median[companion+'_period'] )) for t in tmid_estimates ]) #get corresponding transit number
        nr -= int(nr[-1]/2.) #shift into the middle of the data set
        period_mean, epoch_mean = np.polyfit(nr, tmid_estimates, 1)             
        
        fig, axes = plt.subplots(2,1,figsize=(6,8),tight_layout=True,sharex=True)
        axes[0].plot(nr, tmid_estimates, marker='o', ls='none', color=alles.settings[companion+'_color'], label=companion)
        axes[0].plot(nr, epoch_mean + nr * period_mean, marker='', ls='--', color='grey')
        axes[0].set(ylabel='Transit mid-time')
        axes[0].legend()
        axes[1].plot(nr, tmid_estimates-tmid_linear_predictions, marker='o', ls='none', color=alles.settings[companion+'_color'])
        axes[1].axhline(0, ls='--', color='grey')
        axes[1].set(xlabel='Nr.', ylabel='TTV (min.)')
        fig.savefig(os.path.join(datadir,'ttv_preparation','ttv_preparation_'+companion+'_oc.pdf'), bbox_inches='tight')
        plt.close(fig)
        
        
        #----------------------------------------------------------------------
        #::: compute and output the deviation
        #----------------------------------------------------------------------
        period_dev = np.abs( (period_mean-alles.initial_guess_params_median[companion+'_period'])/alles.initial_guess_params_median[companion+'_period'] )
        epoch_dev = np.abs( (epoch_mean-alles.initial_guess_params_median[companion+'_epoch'])/alles.initial_guess_params_median[companion+'_epoch'] )
        
        print('\nCompanion', companion)
        print('Initial guess for mean period and epoch:')
        print(np.format_float_positional(alles.initial_guess_params_median[companion+'_period']), 
              np.format_float_positional(alles.initial_guess_params_median[companion+'_epoch']))
        print('New estimate for mean period and epoch:')
        print(np.format_float_positional(period_mean,4), 
              np.format_float_positional(epoch_mean,4))
        if (period_dev > 0.01) or (epoch_dev > 0.01):
            print('\n! Consider updating your initial guess to these new estimated mean values.')
            print('\n! If you do, then you must rerun this code.')
        else:
            print('\n! Looks great! You are ready to fit.')


        #----------------------------------------------------------------------
        #::: full lightcurve plot
        #----------------------------------------------------------------------
        #::: plot all data points
        if style=='fullplot':
            axes = fullplot(time_combined, flux_combined, color='silver')
        elif style=='brokenplot':
            axes = brokenplot(time_combined, flux_combined, color='silver')
        elif style=='chunkplot':
            axes = chunkplot(time_combined, flux_combined, color='silver')
        elif style=='monthplot':
            axes = monthplot(time_combined, flux_combined, color='silver')
        elif style=='tessplot':
            axes = tessplot(time_combined, flux_combined, color='silver')
        else:
            raise ValueError("The keyword argument 'style' must be 'fullplot', 'monthplot', or 'tessplot'.")
        
        #::: mark the tranists/eclipses of each photometric companion
        for i, c in enumerate(alles.settings['companions_phot']):
            ind_tr, ind_out = index_transits(time_combined, alles.initial_guess_params_median[c+'_epoch'], alles.initial_guess_params_median[c+'_period'], window)
            if style=='fullplot':
                axes = fullplot(time_combined[ind_tr], flux_combined[ind_tr], color=alles.settings[c+'_color'], ax=axes, label=c)
            elif style=='brokenplot':
                axes = brokenplot(time_combined[ind_tr], flux_combined[ind_tr], color=alles.settings[c+'_color'], bax=axes, label=c)
            elif style=='chunkplot':
                axes = chunkplot(time_combined[ind_tr], flux_combined[ind_tr], color=alles.settings[c+'_color'], axes=axes, label=c)
            elif style=='monthplot':
                axes = monthplot(time_combined[ind_tr], flux_combined[ind_tr], color=alles.settings[c+'_color'], axes=axes, label=c)
            elif style=='tessplot':
                axes = tessplot(time_combined[ind_tr], flux_combined[ind_tr], color=alles.settings[c+'_color'], axes=axes, label=c)
                
        #::: add legend
        axes = np.atleast_1d(axes)
        for i, c in enumerate(alles.settings['companions_phot']):
            axes[0].text(0.02+i*0.02, 0.95, c, color=alles.settings[c+'_color'], ha='left', va='top', transform=axes[0].transAxes, zorder=15)
            
        #::: add vertical lines and numbers
        flux_min = np.nanmin(flux_combined)
        flux_max = np.nanmax(flux_combined)
        for ax in axes:
            for i, tmid in enumerate(alles.data[companion+'_tmid_observed_transits']):
                if (tmid>ax.get_xlim()[0]) & (tmid<ax.get_xlim()[1]):
                    ax.text( tmid, 0.992*flux_min, str(i+1), color=alles.settings[companion+'_color'], ha='center', zorder=12 )  
                    ax.axvline( tmid, color='lightgrey', zorder=11 )
                    ax.set(ylim=[0.99*flux_min, 1.002*flux_max], title='Companion '+companion) 
            
        #::: wrap up
        fname = os.path.join(datadir,'ttv_preparation','ttv_preparation_'+companion+'.jpg')
        fig = plt.gcf()
        fig.savefig(fname, bbox_inches='tight' )  
        plt.close(fig)
                